/**
 * delight.Image
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Image;
private import delight.Math;
private import delight.Util;
private import tango.io.stream.FileStream;

class Image
{
	struct Pixel
	{
		Vector3 color = Vector3(0);
	}
	
	uint width, height;
	Pixel[] pixels;
	
	void initImage(uint w, uint h)
	{
		width = w;
		height = h;
		pixels = new Pixel[w * h];
	}
	
	void setPixel(uint x, uint y, Vector3 c)
	{
		pixels[width * y + x].color = c;
	}
	
	Vector3 getPixel(uint x, uint y)
	{
		return pixels[width * y + x].color;
	}
	
	void saveAsBmp(char[] fileName)
	{
		ubyte[] data = new ubyte[width * height * 4];

		foreach(y ; range(height))
		{
			foreach(x ; range(width))
			{
				static ubyte floatToUbyte(float v)
				{
					return cast(ubyte)clamp!(uint)(cast(uint)(saturate(v) * 256), 0, 255);				
				}

				auto p = &pixels[y * width + x];
				auto i = ((height-1-y) * width + x) * 4;

				data[i + 0] = floatToUbyte(p.color.z);
				data[i + 1] = floatToUbyte(p.color.y);
				data[i + 2] = floatToUbyte(p.color.x);
				data[i + 3] = 0;
			}
		}
		
		struct BmpHeader
		{
		align (1):
			char[2] Magic = "BM";
			uint fileSize;
			uint reserved = 0;
			uint bitmapDataOffset;
		}
		
		struct DibHeaderV3
		{
		align (1):
			uint headerSize = 40;
			uint width, height;
			ushort numColorPlanes = 1;
			ushort bitsPerPixel;
			uint compression = 0;
			uint dataSize;
			uint horzRes = 4096;
			uint vertRes = 4096;
			uint numColorsInPalette = 0;
			uint numImportantColors = 0;
		}
		
		struct Header
		{
		align (1):
			BmpHeader bmp;
			DibHeaderV3 dib;
		}

		static assert(Header.sizeof == 54);

		uint bpp = 32;
		uint dataOffset = Header.sizeof;
		uint dataSize = data.length;

		ubyte[] headerData = new ubyte[dataOffset];
		
		Header * header = cast(Header *)&headerData[0];
		*header = Header.init;

		header.bmp.bitmapDataOffset = dataOffset;
		header.bmp.fileSize = dataOffset + dataSize;
		header.dib.width = width;
		header.dib.height = height;
		header.dib.bitsPerPixel = bpp;
		header.dib.dataSize = dataSize;
		
		scope file = new FileOutput(fileName);
		
		file.write(headerData);
		file.write(data);
	}
}