/**
 * delight.Material
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Material;
import delight.Math;


struct ShadingParams
{
	Vector3 N;  // surface normal
	Vector3 V;  // view direction
	Vector3 R;  // reflect(V, N)
	Vector3 L;  // light direction
	Vector3 CL; // color of light
}

struct Material
{
	Vector3 color = Vector3(1);
	float specular = 1.0f, specularPower = 50.f;
	float fresnel = 0.04f;
	float reflective = 1.0f;
	float refractive = 0.0f;
	float refractionIndex = 1.3f;
	
	Vector3 shade(ShadingParams * params)
	{
		auto d = dot(params.N, params.L);
		
		if (d > 0)
		{
			auto D = color * params.CL * d;
			auto R = reflect(params.V, params.N);
			auto s = pow(saturate(dot(params.L, params.R)), specularPower) * specular;
			auto S = params.CL * s;
			return D + S;
		}
		
		return Vector3(0);
	}
	
	float fresnelReflectance(float fromRefractionIndex, float cosi)
	{
		return saturate(fresnel + pow(1 - cosi, 3.0f));
		/*
		auto etai = fromRefractionIndex;
		auto etat = refractionIndex;
		
		auto sintSq = (1 - cosi * cosi) * (etai * etai) / (etat * etat);
		auto cost = sqrt(1 - sintSq);
		
		auto Rparl = (etat * cosi - etai * cost) / (etat * cosi + etai * cost);
		auto Rperp = (etai * cosi - etat * cost) / (etai * cosi + etat * cost);
		return (Rparl * Rparl + Rperp * Rperp) / 2.f;
		*/
	}
	
	Vector3 refract(Vector3 v, Vector3 N, float fromRefractionIndex)
	{
		auto etai = fromRefractionIndex;
		auto etat = refractionIndex;
		
		auto cosi = -dot(v, N);
		auto sintSq = (1 - cosi * cosi) * (etai * etai) / (etat * etat);
		auto cost = sqrt(1 - sintSq);
		auto sint = sqrt(sintSq);
		
		auto ux = normalize(orthogonalize(v, N));
		
		return ux * sint - N * cost;		
	}
}

class MaterialObject
{
	char[] name;
	Material material;
	
	this(char[] n, Material m)
	{
		name = n;
		material = m;
	}
}
