/**
 * delight.Ray
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Ray;

private import delight.Scene;
private import delight.Math;
private import delight.Util;
private import Math = tango.math.Math;
private import tango.io.Stdout;
private import tango.text.convert.Format;


struct Ray
{
	static const defaultMinDist = 0.000000001f;
	static const defaultMaxDist = 100000.f;
	
	Vector3 origin = Vector3(0);
	Vector3 direction = Vector3(1,0,0);
	float minDistance = defaultMinDist;
	float maxDistance = 100000;
	
	static Ray opCall(Vector3 o, Vector3 d, float minD = defaultMinDist, float maxD = defaultMaxDist)
	{
		Ray ray = { o, d, minD, maxD };
		return ray;
	}
	
	Vector3 position(float t) { return origin + direction * t; } 
	
	bool insideRange(float d) { return d > minDistance && d < maxDistance; }

	bool intersectsRange(float a, float b) { return b > minDistance && a < maxDistance; }
	
	bool intersects(AABB aabb)
	{
		// Ray-AABB Intersection
		
		struct F2
		{
			float a, b;
			
			F2 opNeg()
			{
				return F2(-a, -b);
			}
			
			float toFloat() { return a / b; }

			int opCmp(F2 e2)
			{
				float r = a * e2.b - e2.a * b;
				if (r < 0) return -1;
				if (r > 0) return 1;
				return 0;
			}
			
			char[] toString() { return Format.convert("({0},{1})", a, b); }
		}

		F2 f2left(uint i)  { return F2(aabb.v1[i] - origin[i], direction[i]); }
		F2 f2right(uint i) { return F2(aabb.v2[i] - origin[i], direction[i]); }
		
		F2[3] left  = void;
		F2[3] right = void;
		
		foreach( i ; range(3) )
		{
			if (direction[i] < 0)
			{
				left[i] = -f2right(i);
				right[i] = -f2left(i);
			}
			else
			{
				left[i] = f2left(i);
				right[i] = f2right(i);
			}
		}
		
		F2 maxLeft = Math.max(left[0], left[1], left[2]);
		F2 minRight = Math.min(right[0], right[1], right[2]);

		if (maxLeft <= minRight)
		{
			return intersectsRange(maxLeft.toFloat, minRight.toFloat);
		}
		
		return false;
	}
	
	char[] toString()
	{
		return Format.convert("({0}, {1}, {2}, {3})", origin, direction, minDistance, maxDistance);
	}
}

struct RayContact
{
	Ray ray;
	SceneObject object;
	float distance = 200000;
	Vector3 normal;
	
	Vector3 contactPosition() { return ray.position(distance); }
}
